@@ -1,5 +1,5 @@
 # Maildir folder support
-# Copyright (C) 2002-2015 John Goerzen & contributors
+# Copyright (C) 2002-2016 John Goerzen & contributors.
 #
 #    This program is free software; you can redistribute it and/or modify
 #    it under the terms of the GNU General Public License as published by
@@ -15,66 +15,69 @@
 #    along with this program; if not, write to the Free Software
 #    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 
+import errno
 import socket
 import time
 import re
 import os
 from sys import exc_info
-from .Base import BaseFolder
 from threading import Lock
-try:
-    from hashlib import md5
-except ImportError:
-    from md5 import md5
-try: # python 2.6 has set() built in
-    set
-except NameError:
-    from sets import Set as set
-
-from offlineimap import OfflineImapError, emailutil
+from hashlib import md5
+from offlineimap import OfflineImapError
+from .Base import BaseFolder
+from email.errors import NoBoundaryInMultipartDefect
 
 # Find the UID in a message filename
-re_uidmatch = re.compile(',U=(\d+)')
+re_uidmatch = re.compile(r',U=(\d+)')
 # Find a numeric timestamp in a string (filename prefix)
-re_timestampmatch = re.compile('(\d+)');
+re_timestampmatch = re.compile(r'(\d+)')
 
 timehash = {}
 timelock = Lock()
 
+
 def _gettimeseq(date=None):
     global timehash, timelock
     timelock.acquire()
     try:
         if date is None:
-            date = long(time.time())
-        if timehash.has_key(date):
+            date = int(time.time())
+        if date in timehash:
             timehash[date] += 1
         else:
             timehash[date] = 0
-        return (date, timehash[date])
+        return date, timehash[date]
     finally:
         timelock.release()
 
+
 class MaildirFolder(BaseFolder):
     def __init__(self, root, name, sep, repository):
-        self.sep = sep # needs to be set before super().__init__
+        self.sep = sep  # needs to be set before super().__init__
         super(MaildirFolder, self).__init__(name, repository)
-        self.dofsync = self.config.getdefaultboolean("general", "fsync", True)
         self.root = root
-        self.messagelist = {}
         # check if we should use a different infosep to support Win file systems
         self.wincompatible = self.config.getdefaultboolean(
-            "Account "+self.accountname, "maildir-windows-compatible", False)
+            "Account " + self.accountname, "maildir-windows-compatible", False)
         self.infosep = '!' if self.wincompatible else ':'
         """infosep is the separator between maildir name and flag appendix"""
-        self.re_flagmatch = re.compile('%s2,(\w*)'% self.infosep)
-        #self.ui is set in BaseFolder.init()
+        self.re_flagmatch = re.compile(r'%s2,(\w*)' % self.infosep)
+        # self.ui is set in BaseFolder.init()
         # Everything up to the first comma or colon (or ! if Windows):
-        self.re_prefixmatch = re.compile('([^'+ self.infosep + ',]*)')
+        self.re_prefixmatch = re.compile('([^' + self.infosep + ',]*)')
         # folder's md, so we can match with recorded file md5 for validity.
-        self._foldermd5 = md5(self.getvisiblename()).hexdigest()
+        self._foldermd5 = md5(self.getvisiblename().encode('utf-8')).hexdigest()
         # Cache the full folder path, as we use getfullname() very often.
         self._fullname = os.path.join(self.getroot(), self.getname())
+        # Modification time from 'Date' header.
+        utime_from_header_global = self.config.getdefaultboolean(
+            "general", "utime_from_header", False)
+        self._utime_from_header = self.config.getdefaultboolean(
+            self.repoconfname, "utime_from_header", utime_from_header_global)
+        # What do we substitute pathname separator in names (if any)
+        self.sep_subst = '-'
+        if os.path.sep == self.sep_subst:
+            self.sep_subst = '_'
 
     # Interface from BaseFolder
     def getfullname(self):
@@ -98,8 +101,8 @@ class MaildirFolder(BaseFolder):
         if not timestampmatch:
             return True
         timestampstr = timestampmatch.group()
-        timestamplong = long(timestampstr)
-        if(timestamplong < time.mktime(date)):
+        timestamplong = int(timestampstr)
+        if timestamplong < time.mktime(date):
             return False
         else:
             return True
@@ -125,14 +128,21 @@ class MaildirFolder(BaseFolder):
         prefixmatch = self.re_prefixmatch.match(filename)
         if prefixmatch:
             prefix = prefixmatch.group(1)
-        folderstr = ',FMD5=%s'% self._foldermd5
+        folderstr = ',FMD5=%s' % self._foldermd5
         foldermatch = folderstr in filename
         # If there was no folder MD5 specified, or if it mismatches,
         # assume it is a foreign (new) message and ret: uid, fmd5 = None, None
+
+        # XXX: This is wrong behaviour: if FMD5 is missing or mismatches, assume
+        # the mail is new and **fix UID to None** to avoid any conflict.
+
+        # XXX: If UID is missing, I have no idea what FMD5 can do. Should be
+        # fixed to None in this case, too.
+
         if foldermatch:
             uidmatch = re_uidmatch.search(filename)
             if uidmatch:
-                uid = long(uidmatch.group(1))
+                uid = int(uidmatch.group(1))
         flagmatch = self.re_flagmatch.search(filename)
         if flagmatch:
             flags = set((c for c in flagmatch.group(1)))
@@ -146,8 +156,9 @@ class MaildirFolder(BaseFolder):
         This handles the edge cases where the date is much earlier than messages
         with similar UID's (e.g. the UID was reassigned much later).
 
-        Maildir flags are: R (replied) S (seen) T (trashed) D (draft) F
-        (flagged), plus lower-case letters for custom flags.
+        Maildir flags are:
+            D (draft) F (flagged) R (replied) S (seen) T (trashed),
+        plus lower-case letters for custom flags.
         :returns: dict that can be used as self.messagelist.
         """
 
@@ -155,7 +166,7 @@ class MaildirFolder(BaseFolder):
 
         retval = {}
         files = []
-        nouidcounter = -1          # Messages without UIDs get negative UIDs.
+        nouidcounter = -1  # Messages without UIDs get negative UIDs.
         for dirannex in ['new', 'cur']:
             fulldirname = os.path.join(self.getfullname(), dirannex)
             files.extend((dirannex, filename) for
@@ -164,29 +175,28 @@ class MaildirFolder(BaseFolder):
         date_excludees = {}
         for dirannex, filename in files:
             if filename.startswith('.'):
-                continue # Ignore dot files.
+                continue  # Ignore dot files.
             # We store just dirannex and filename, ie 'cur/123...'
             filepath = os.path.join(dirannex, filename)
             # Check maxsize if this message should be considered.
-            if maxsize and (os.path.getsize(os.path.join(
-                        self.getfullname(), filepath)) > maxsize):
+            if maxsize and (os.path.getsize(
+                    os.path.join(self.getfullname(), filepath)) > maxsize):
                 continue
 
-            (prefix, uid, fmd5, flags) = self._parse_filename(filename)
-            if uid is None: # Assign negative uid to upload it.
+            prefix, uid, fmd5, flags = self._parse_filename(filename)
+            if uid is None:  # Assign negative uid to upload it.
                 uid = nouidcounter
                 nouidcounter -= 1
-            else:                       # It comes from our folder.
+            else:  # It comes from our folder.
                 uidmatch = re_uidmatch.search(filename)
-                uid = None
                 if not uidmatch:
                     uid = nouidcounter
                     nouidcounter -= 1
                 else:
-                    uid = long(uidmatch.group(1))
-            if min_uid != None and uid > 0 and uid < min_uid:
+                    uid = int(uidmatch.group(1))
+            if min_uid is not None and uid > 0 and uid < min_uid:
                 continue
-            if min_date != None and not self._iswithintime(filename, min_date):
+            if min_date is not None and not self._iswithintime(filename, min_date):
                 # Keep track of messages outside of the time limit, because they
                 # still might have UID > min(UIDs of within-min_date). We hit
                 # this case for maxage if any message had a known/valid datetime
@@ -203,12 +213,12 @@ class MaildirFolder(BaseFolder):
                 retval[uid] = self.msglist_item_initializer(uid)
                 retval[uid]['flags'] = flags
                 retval[uid]['filename'] = filepath
-        if min_date != None:
+        if min_date is not None:
             # Re-include messages with high enough uid's.
-            positive_uids = filter(lambda uid: uid > 0, retval)
+            positive_uids = [uid for uid in retval if uid > 0]
             if positive_uids:
                 min_uid = min(positive_uids)
-                for uid in date_excludees.keys():
+                for uid in list(date_excludees.keys()):
                     if uid > min_uid:
                         # This message was originally excluded because of
                         # its date. It is re-included now because we want all
@@ -226,12 +236,11 @@ class MaildirFolder(BaseFolder):
                 sorted(statusfolder.getmessageuidlist()):
             return True
         # Also check for flag changes, it's quick on a Maildir.
-        for (uid, message) in self.getmessagelist().iteritems():
+        for (uid, message) in list(self.getmessagelist().items()):
             if message['flags'] != statusfolder.getmessageflags(uid):
                 return True
         return False  # Nope, nothing changed.
 
-
     # Interface from BaseFolder
     def msglist_item_initializer(self, uid):
         return {'flags': set(), 'filename': '/no-dir/no-such-file/'}
@@ -241,25 +250,47 @@ class MaildirFolder(BaseFolder):
         if self.ismessagelistempty():
             self.ui.loadmessagelist(self.repository, self)
             self.messagelist = self._scanfolder(min_date=min_date,
-                min_uid=min_uid)
+                                                min_uid=min_uid)
             self.ui.messagelistloaded(self.repository, self, self.getmessagecount())
 
-    # Interface from BaseFolder
-    def getmessagelist(self):
-        return self.messagelist
-
     # Interface from BaseFolder
     def getmessage(self, uid):
-        """Return the content of the message."""
+        """Returns an email message object."""
 
         filename = self.messagelist[uid]['filename']
         filepath = os.path.join(self.getfullname(), filename)
-        file = open(filepath, 'rt')
-        retval = file.read()
-        file.close()
-        #TODO: WHY are we replacing \r\n with \n here? And why do we
-        #      read it as text?
-        return retval.replace("\r\n", "\n")
+        fd = open(filepath, 'rb')
+        _fd_bytes = fd.read()
+        fd.close()
+        try: retval = self.parser['8bit'].parsebytes(_fd_bytes)
+        except:
+            err = exc_info()
+            msg_id = self._extract_message_id(_fd_bytes)[0].decode('ascii',errors='surrogateescape')
+            raise OfflineImapError(
+                "Exception parsing message with ID ({}) from file ({}).\n {}: {}".format(
+                    msg_id, filename, err[0].__name__, err[1]),
+                OfflineImapError.ERROR.MESSAGE)
+        if len(retval.defects) > 0:
+            # We don't automatically apply fixes as to attempt to preserve the original message
+            self.ui.warn("UID {} has defects: {}".format(uid, retval.defects))
+            if any(isinstance(defect, NoBoundaryInMultipartDefect) for defect in retval.defects):
+                # (Hopefully) Rare defect from a broken client where multipart boundary is
+                # not properly quoted.  Attempt to solve by fixing the boundary and parsing
+                self.ui.warn(" ... applying multipart boundary fix.")
+                retval = self.parser['8bit'].parsebytes(self._quote_boundary_fix(_fd_bytes))
+            try:
+                # See if the defects after fixes are preventing us from obtaining bytes
+                _ = retval.as_bytes(policy=self.policy['8bit'])
+            except UnicodeEncodeError as err:
+                # Unknown issue which is causing failure of as_bytes()
+                msg_id = self.getmessageheader(retval, "message-id")
+                if msg_id is None:
+                    msg_id = '<unknown-message-id>'
+                raise OfflineImapError(
+                        "UID {} ({}) has defects preventing it from being processed!\n  {}: {}".format(
+                            uid, msg_id, type(err).__name__, err),
+                        OfflineImapError.ERROR.MESSAGE)
+        return retval
 
     # Interface from BaseFolder
     def getmessagetime(self, uid):
@@ -267,30 +298,38 @@ class MaildirFolder(BaseFolder):
         filepath = os.path.join(self.getfullname(), filename)
         return os.path.getmtime(filepath)
 
-    def new_message_filename(self, uid, flags=set(), date=None):
+    def new_message_filename(self, uid, flags=None, date=None):
         """Creates a new unique Maildir filename
 
         :param uid: The UID`None`, or a set of maildir flags
         :param flags: A set of maildir flags
+        :param flags: (optional) Date
         :returns: String containing unique message filename"""
 
+        if flags is None:
+            flags = set()
+
         timeval, timeseq = _gettimeseq(date)
-        return '%d_%d.%d.%s,U=%d,FMD5=%s%s2,%s'% \
-            (timeval, timeseq, os.getpid(), socket.gethostname(),
-            uid, self._foldermd5, self.infosep, ''.join(sorted(flags)))
+        uniq_name = '%d_%d.%d.%s,U=%d,FMD5=%s%s2,%s' % \
+                    (timeval, timeseq, os.getpid(), socket.gethostname(),
+                     uid, self._foldermd5, self.infosep, ''.join(sorted(flags)))
+        return uniq_name.replace(os.path.sep, self.sep_subst)
 
-
-    def save_to_tmp_file(self, filename, content):
-        """Saves given content to the named temporary file in the
+    def save_to_tmp_file(self, filename, msg, policy=None):
+        """Saves given message to the named temporary file in the
         'tmp' subdirectory of $CWD.
 
         Arguments:
         - filename: name of the temporary file;
-        - content: data to be saved.
+        - msg: Email message object
 
         Returns: relative path to the temporary file
         that was created."""
 
+        if policy is None:
+            output_policy = self.policy['8bit']
+        else:
+            output_policy = policy
         tmpname = os.path.join('tmp', filename)
         # Open file and write it out.
         # XXX: why do we need to loop 7 times?
@@ -299,39 +338,41 @@ class MaildirFolder(BaseFolder):
             tries = tries - 1
             try:
                 fd = os.open(os.path.join(self.getfullname(), tmpname),
-                             os.O_EXCL|os.O_CREAT|os.O_WRONLY, 0o666)
+                             os.O_EXCL | os.O_CREAT | os.O_WRONLY, 0o666)
                 break
             except OSError as e:
                 if not hasattr(e, 'EEXIST'):
                     raise
-                if e.errno == e.EEXIST:
+                if e.errno == errno.EEXIST:
                     if tries:
                         time.sleep(0.23)
                         continue
                     severity = OfflineImapError.ERROR.MESSAGE
-                    raise OfflineImapError("Unique filename %s already exists."%
-                        filename, severity), None, exc_info()[2]
+                    raise OfflineImapError(
+                        "Unique filename %s already exists." %
+                        filename, severity,
+                        exc_info()[2])
                 else:
                     raise
 
-        fd = os.fdopen(fd, 'wt')
-        fd.write(content)
+        fd = os.fdopen(fd, 'wb')
+        fd.write(msg.as_bytes(policy=output_policy))
         # Make sure the data hits the disk.
         fd.flush()
-        if self.dofsync:
+        if self.dofsync():
             os.fsync(fd)
         fd.close()
 
         return tmpname
 
-
     # Interface from BaseFolder
-    def savemessage(self, uid, content, flags, rtime):
+    def savemessage(self, uid, msg, flags, rtime):
         """Writes a new message, with the specified uid.
 
         See folder/Base for detail. Note that savemessage() does not
         check against dryrun settings, so you need to ensure that
         savemessage is never called in a dryrun mode."""
+
         # This function only ever saves to tmp/,
         # but it calls savemessageflags() to actually save to cur/ or new/.
         self.ui.savemessage('maildir', uid, flags, self)
@@ -344,51 +385,46 @@ class MaildirFolder(BaseFolder):
             self.savemessageflags(uid, flags)
             return uid
 
-        # Otherwise, save the message in tmp/ and then call savemessageflags()
-        # to give it a permanent home.
-        tmpdir = os.path.join(self.getfullname(), 'tmp')
-
-        # use the mail timestamp given by either Date or Delivery-date mail
+        # Use the mail timestamp given by either Date or Delivery-date mail
         # headers.
         message_timestamp = None
-        if self._filename_use_mail_timestamp:
+        if self._filename_use_mail_timestamp is not False:
             try:
-                message_timestamp = emailutil.get_message_date(content, 'Date')
+                message_timestamp = self.get_message_date(msg, 'Date')
                 if message_timestamp is None:
                     # Give a try with Delivery-date
-                    date = emailutil.get_message_date(content, 'Delivery-date')
-            except:
-                # This should never happen
-                from email.Parser import Parser
+                    message_timestamp = self.get_message_date(
+                        msg, 'Delivery-date')
+            except Exception as e:
+                # Extracting the date has failed for some reason, such as it
+                # being in an invalid format.
                 from offlineimap.ui import getglobalui
-                datestr = Parser().parsestr(content, True).get("Date")
                 ui = getglobalui()
-                ui.warn("UID %d has invalid date %s: %s\n"
-                    "Not using message timestamp as file prefix" % (uid, datestr, e))
-                # No need to check if date is None here since it would
-                # be overridden by _gettimeseq.
+                ui.warn("UID %d has invalid date: %s\n"
+                        "Not using message timestamp as file prefix" % (uid, e))
+                # No need to check if message_timestamp is None here since it
+                # would be overridden by _gettimeseq.
         messagename = self.new_message_filename(uid, flags, date=message_timestamp)
-        tmpname = self.save_to_tmp_file(messagename, content)
+        tmpname = self.save_to_tmp_file(messagename, msg)
 
-        if self.utime_from_header:
+        if self._utime_from_header is True:
             try:
-                date = emailutil.get_message_date(content, 'Date')
+                date = self.get_message_date(msg, 'Date')
                 if date is not None:
                     os.utime(os.path.join(self.getfullname(), tmpname),
-                        (date, date))
-            # In case date is wrongly so far into the future as to be > max int32
+                             (date, date))
             except Exception as e:
-                from email.Parser import Parser
+                # Extracting the date has failed for some reason, such as it
+                # being in an invalid format.
                 from offlineimap.ui import getglobalui
-                datestr = Parser().parsestr(content, True).get("Date")
                 ui = getglobalui()
-                ui.warn("UID %d has invalid date %s: %s\n"
-                    "Not changing file modification time" % (uid, datestr, e))
+                ui.warn("UID %d has invalid date: %s\n"
+                        "Not changing file modification time" % (uid, e))
 
         self.messagelist[uid] = self.msglist_item_initializer(uid)
         self.messagelist[uid]['flags'] = flags
         self.messagelist[uid]['filename'] = tmpname
-        # savemessageflags moves msg to 'cur' or 'new' as appropriate
+        # savemessageflags moves msg to 'cur' or 'new' as appropriate.
         self.savemessageflags(uid, flags)
         self.ui.debug('maildir', 'savemessage: returning uid %d' % uid)
         return uid
@@ -420,20 +456,21 @@ class MaildirFolder(BaseFolder):
             # off existing infostring
             infomatch = self.re_flagmatch.search(filename)
             if infomatch:
-                filename = filename[:-len(infomatch.group())] #strip off
-            infostr = '%s2,%s'% (self.infosep, ''.join(sorted(flags)))
+                filename = filename[:-len(infomatch.group())]  # strip off
+            infostr = '%s2,%s' % (self.infosep, ''.join(sorted(flags)))
             filename += infostr
 
         newfilename = os.path.join(dir_prefix, filename)
-        if (newfilename != oldfilename):
+        if newfilename != oldfilename:
             try:
                 os.rename(os.path.join(self.getfullname(), oldfilename),
                           os.path.join(self.getfullname(), newfilename))
             except OSError as e:
-                raise OfflineImapError("Can't rename file '%s' to '%s': %s" % (
-                                       oldfilename, newfilename, e[1]),
-                                       OfflineImapError.ERROR.FOLDER), \
-                      None, exc_info()[2]
+                raise OfflineImapError(
+                    "Can't rename file '%s' to '%s': %s" %
+                    (oldfilename, newfilename, e.errno),
+                    OfflineImapError.ERROR.FOLDER,
+                    exc_info()[2])
 
             self.messagelist[uid]['flags'] = flags
             self.messagelist[uid]['filename'] = newfilename
@@ -443,20 +480,25 @@ class MaildirFolder(BaseFolder):
         """Change the message from existing uid to new_uid
 
         This will not update the statusfolder UID, you need to do that yourself.
+        :param uid: Message UID
         :param new_uid: (optional) If given, the old UID will be changed
                         to a new UID. The Maildir backend can implement this as
                         an efficient rename.
         """
 
-        if not uid in self.messagelist:
-            raise OfflineImapError("Cannot change unknown Maildir UID %s"% uid)
-        if uid == new_uid: return
+        if uid not in self.messagelist:
+            raise OfflineImapError("Cannot change unknown Maildir UID %s" % uid,
+                                   OfflineImapError.ERROR.MESSAGE)
+        if uid == new_uid:
+            return
 
         oldfilename = self.messagelist[uid]['filename']
         dir_prefix, filename = os.path.split(oldfilename)
         flags = self.getmessageflags(uid)
+        # TODO: we aren't keeping the prefix timestamp so we don't honor the
+        # filename_use_mail_timestamp configuration option.
         newfilename = os.path.join(dir_prefix,
-          self.new_message_filename(new_uid, flags))
+                                   self.new_message_filename(new_uid, flags))
         os.rename(os.path.join(self.getfullname(), oldfilename),
                   os.path.join(self.getfullname(), newfilename))
         self.messagelist[new_uid] = self.messagelist[uid]
@@ -479,23 +521,22 @@ class MaildirFolder(BaseFolder):
         except OSError:
             # Can't find the file -- maybe already deleted?
             newmsglist = self._scanfolder()
-            if uid in newmsglist:       # Nope, try new filename.
+            if uid in newmsglist:  # Nope, try new filename.
                 filename = newmsglist[uid]['filename']
                 filepath = os.path.join(self.getfullname(), filename)
                 os.unlink(filepath)
             # Yep -- return.
-        del(self.messagelist[uid])
+        del (self.messagelist[uid])
 
     def migratefmd5(self, dryrun=False):
         """Migrate FMD5 hashes from versions prior to 6.3.5
 
         :param dryrun: Run in dry run mode
-        :type fix: Boolean
         :return: None
         """
         oldfmd5 = md5(self.name).hexdigest()
         msglist = self._scanfolder()
-        for mkey, mvalue in msglist.iteritems():
+        for mkey, mvalue in list(msglist.items()):
             filename = os.path.join(self.getfullname(), mvalue['filename'])
             match = re.search("FMD5=([a-fA-F0-9]+)", filename)
             if match is None:
@@ -512,9 +553,11 @@ class MaildirFolder(BaseFolder):
                         os.rename(filename, newfilename)
                     except OSError as e:
                         raise OfflineImapError(
-                            "Can't rename file '%s' to '%s': %s" % (
-                                filename, newfilename, e[1]),
-                            OfflineImapError.ERROR.FOLDER), None, exc_info()[2]
+                            "Can't rename file '%s' to '%s': %s" %
+                            (filename, newfilename, e.errno),
+                            OfflineImapError.ERROR.FOLDER,
+                            exc_info()[2])
+
             elif match.group(1) != self._foldermd5:
                 self.ui.warn(("Inconsistent FMD5 for file `%s':"
                               " Neither `%s' nor `%s' found")
